@@ -56,7 +56,7 @@ static PyObject *Error, *Warning, *InterfaceError,
 	*DatabaseError, *InternalError, *OperationalError, *ProgrammingError,
 	*IntegrityError, *DataError, *NotSupportedError;
 
-static const char *PyPgVersion = "0.9.2";
+static const char *PyPgVersion = "0.9.3";
 static char pg__doc__[] = "Simple Python interface to PostgreSQL DB";
 
 /* default values */
@@ -417,6 +417,7 @@ static PyObject *_pgsource_postexec(pgsourceobject * self)
 	    self->result_type = RESULT_DQL;
 	    self->max_row = PQntuples(self->last_result);
 	    self->num_fields = PQnfields(self->last_result);
+	    self->current_row = 0;
 	    Py_INCREF(Py_None);
 	    return Py_None;
 	case PGRES_COMMAND_OK:	/* other requests */
@@ -573,44 +574,43 @@ static pgparams *_pgsource_getparams(PyObject *params)
 	    ret->paramValues[i] = dblstr;
 	    ret->mustFree[i] = 1;
 	} else if (param == Py_None) {
-	    ret->paramTypes[i] = VOIDOID;
-	} else {
+	    ret->paramTypes[i] = 0;
+	    ret->paramValues[i] = 0;
+	} else { /* this is an object that hopefully we can str() */
 	    PyObject *o = NULL;
 	    PyObject *str = NULL;
+	    int len;
+	    char *value;
+
+	    /* is this an object that needs to be treated as a binary one? */
+	    if (PyObject_HasAttrString(param, "__binary__"))
+		ret->paramFormats[i] = 1;
+	    /* try to quote the object */
 	    if (PyObject_HasAttrString(param, "__pgquote__"))
 		o = PyObject_GetAttrString(param, "__pgquote__");
 	    else if (PyObject_HasAttrString(param, "__quote__"))
 		o = PyObject_GetAttrString(param, "__quote__");
-	    else if (PyObject_HasAttrString(param, "_quote"))
-		o = PyObject_GetAttrString(param, "_quote");
 
-	    if (o != NULL && PyCallable_Check(o)) {
-		int len;
-		char *value;
+	    if (o != NULL && PyCallable_Check(o))
 		str = PyObject_CallObject(o, NULL);
-		PyString_AsStringAndSize(str, &value, &len);
-		ret->paramValues[i] = calloc(1, len+1);
-		if (ret->paramValues[i] == NULL) {
-		    Py_DECREF(param);
-		    Py_DECREF(str);
-		    _pgsource_freeparams(ret);
-		    PyErr_SetString(ProgrammingError, "out of memory binding paramaters");
-		    return NULL;
-		}
-		memcpy(ret->paramValues[i], value, len);
-		ret->mustFree[i] = 1;
-		ret->paramTypes[i] = TEXTOID;
-		ret->paramLengths[i] = len;
+	    else /* hope for the best */
+		str = PyObject_Str(param);
+
+	    PyString_AsStringAndSize(str, &value, &len);
+	    ret->paramValues[i] = calloc(1, len+1);
+	    if (ret->paramValues[i] == NULL) {
+		Py_XDECREF(o);
 		Py_DECREF(str);
-	    } else {
-		str = (PyObject *)PyString_FromString("can not bind param type: ");
-		PyString_ConcatAndDel(&str, PyObject_Repr(param));
-		PyErr_SetObject(ProgrammingError, str);
 		Py_DECREF(param);
-		Py_DECREF(str);
 		_pgsource_freeparams(ret);
+		PyErr_SetString(ProgrammingError, "out of memory binding paramaters");
 		return NULL;
 	    }
+	    ret->mustFree[i] = 1;
+	    memcpy(ret->paramValues[i], value, len);
+	    ret->paramLengths[i] = len;
+	    Py_XDECREF(o);
+	    Py_DECREF(str);
 	}
 	Py_DECREF(param);
     }
@@ -693,7 +693,9 @@ pgsource_execute(pgsourceobject *self, PyObject * args)
 	    PyErr_SetString(ProgrammingError, "execute with parameters requires params as a sequence");
 	    return NULL;
 	}
-    binds = _pgsource_getparams(params);
+
+    if ((binds = _pgsource_getparams(params)) == NULL)
+	return NULL;
 
     /* now run the query */
     Py_BEGIN_ALLOW_THREADS ;
@@ -714,7 +716,7 @@ pgsource_execute(pgsourceobject *self, PyObject * args)
 					 query,
 					 binds->nParams,
 					 binds->paramTypes,
-					 (const char **) binds->paramValues,
+					 (const char **)binds->paramValues,
 					 binds->paramLengths,
 					 binds->paramFormats,
 					 0);
@@ -1123,7 +1125,7 @@ pgsource_fetchall(pgsourceobject * self, PyObject * args)
     }
 
     /* stores result in presized list for efficiency */
-    reslist = PyList_New(self->max_row);
+    reslist = PyList_New(self->max_row - self->current_row);
 
     /* return the remaining rows that have not been "extracted" yet */
     for (row = self->current_row; row < self->max_row; row++) {
@@ -1132,7 +1134,7 @@ pgsource_fetchall(pgsourceobject * self, PyObject * args)
 	    Py_DECREF(reslist);
 	    return NULL;
 	}
-	PyList_SET_ITEM(reslist, row, rowtuple);
+	PyList_SET_ITEM(reslist, row - self->current_row, rowtuple);
     }
     /* mark all rows returned */
     self->current_row = self->max_row + 1;
@@ -1166,7 +1168,7 @@ pgsource_fetchalldict(pgsourceobject * self, PyObject * args)
     }
 
     /* stores result in presized list for efficiency */
-    reslist = PyList_New(self->max_row);
+    reslist = PyList_New(self->max_row - self->current_row);
 
     /* return the remaining rows that have not been "extracted" yet */
     for (row = self->current_row; row < self->max_row; row++) {
@@ -1175,7 +1177,7 @@ pgsource_fetchalldict(pgsourceobject * self, PyObject * args)
 	    Py_DECREF(reslist);
 	    return NULL;
 	}
-	PyList_SET_ITEM(reslist, row, rowdict);
+	PyList_SET_ITEM(reslist, row - self->current_row, rowdict);
     }
     /* mark all rows returned */
     self->current_row = self->max_row + 1;
