@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 #
-# pgsql.py version 0.9.2
+# pgsql.py version 0.9.3
 #
 # Updated by Cristian Gafton to use the bind parameters
 # and adapt to the new interface provided by the C bindings
@@ -45,8 +45,11 @@ except ImportError: # otherwise use standard datetime module
     from datetime import datetime as DateTime, \
          timedelta as TimeDelta, datetime as DateTimeType
 
+import weakref
+__CONNLIST = []
+
 ### module constants
-version = "0.9.2"
+version = "0.9.3"
 
 # compliant with DB SIG 2.0
 apilevel = '2.0'
@@ -87,8 +90,25 @@ class Cursor(object):
     def arraysize(self, size):
         self.arraysize = size
 
+    def _start(self, operation = None):
+        transaction = self.transaction
+        if transaction == TRANS_UNKNOWN:
+            raise DatabaseError("Invalid/Unknown database connection")
+        if transaction == TRANS_INERROR:
+            raise ProgrammingError("Invalid transaction state. "
+                                   "Exception not cleared by COMMIT or ROLLBACK")
+        # do we need to start a transaction?
+        if transaction in [TRANS_INTRANS, TRANS_ACTIVE]:
+            return
+        if operation is None:
+            return
+        op = operation.strip()[:6].lower()
+        if op in ["insert", "update", "delete"]:
+            self._source.execute("START TRANSACTION")
+
     # if parameters are passed in, we'll attempt to bind them
     def execute(self, operation, *params):
+        self._start(operation)
         if not len(params):
             ret = self._source.execute(operation)
         elif isinstance(params[0], (list, tuple)):
@@ -100,6 +120,7 @@ class Cursor(object):
         return self
 
     def executemany(self, operation, param_seq):
+        self._start(operation)
         ret = self._source.executemany(operation, param_seq)
         return self
 
@@ -114,6 +135,16 @@ class Cursor(object):
             size = self.arraysize
         return [self._source.fetchone() for x in range(size)]
 
+    # extensions
+    def fetchone_dict(self):
+        return self._source.fetchonedict()
+    def fetchall_dict(self):
+        return self._source.fetchalldict()
+    def fetchmany_dict(self, size = None):
+        if size == None:
+            size = self.arraysize
+        return [self._source.fetchonedict() for x in range(size)]
+
     def nextset(self):
         raise NotSupportedError, "nextset() is not supported"
 
@@ -138,7 +169,8 @@ class Cursor(object):
                         "arraysize", "resulttype", "rownumber",
                         "oidstatus", "valid"]):
             return getattr(self._source, name)
-        elif name in set(["escape_string", "escape_bytea", "unescape_bytea"]):
+        elif name in set(["escape_string", "escape_bytea", "unescape_bytea",
+                          "transaction"]):
             return getattr(self._source.connection, name)
         elif self.__dict__.has_key(name):
             return self.__dict__[name]
@@ -152,8 +184,17 @@ class Cursor(object):
 
 # A cursor class for prepared statements
 class PreparedCursor(Cursor):
+    def __init__(self, source):
+        Cursor.__init__(self, source)
+        # make sure we're always in a transaction when we're preparing statements
+        if self.transaction == TRANS_IDLE:
+            self._source.connection.execute("START TRANSACTION")
+
     # we require parameters since we've already bound a query
     def execute(self, *params):
+        #self._start()
+        if self.transaction == TRANS_IDLE:
+            self._source.connection.execute("START TRANSACTION")
         if not len(params):
             ret = self._source.execute(None)
         elif isinstance(params[0], (tuple, list)):
@@ -164,6 +205,9 @@ class PreparedCursor(Cursor):
             return ret
         return self
     def executemany(self, param_seq):
+        #self._start()
+        if self.transaction == TRANS_IDLE:
+            self._source.connection.execute("START TRANSACTION")
         ret = self._source.executemany(param_seq)
         return self
 
@@ -173,16 +217,14 @@ class IterCursor(Cursor):
         Cursor.__init__(self, source)
         self.active = 0
         # we need a fairly random name for our cursor executions
-        self.name = "p%did%s" % (os.getpid(), hex(abs(id(source))))
+        self.name = "c%ss%s" % (hex(abs(id(self))), hex(abs(id(source))))
 
     def execute(self, query, *params):
         query = query.strip()
         if not query.upper().startswith("SELECT"):
             return Cursor.execute(self, query, *params)
-        # we have a select query, open up the cursor
-        if self.active:
-            self._source.execute("CLOSE %s" % self.name)
-            self.active = 0
+        # we have a select query
+        self.close()
         # open up this cursor for select
         query = "DECLARE %s NO SCROLL CURSOR WITH HOLD FOR\n%s" %(
             self.name, query)
@@ -195,24 +237,44 @@ class IterCursor(Cursor):
         self.active = 1
         return ret
 
-    def fetchone(self):
+    def __fetchone(self):
         # if this is not an active cursor, passthrough
         if self.active:
             self._source.query("FETCH NEXT FROM %s" % self.name)
+    def fetchone(self):
+        self.__fetchone()
         return Cursor.fetchone(self)
+    def fetchone_dict(self):
+        self.__fetchone()
+        return Cursor.fetchone_dict(self)
 
-    def fetchall(self):
+    def __fetchall(self):
         if self.active:
             self._source.execute("FETCH ALL FROM %s" % self.name)
+    def fetchall(self):
+        self.__fetchall()
         return Cursor.fetchall(self)
+    def fetchall_dict(self):
+        self.__fetchall()
+        return Cursor.fetchall_dict(self)
 
-    def fetchmany(self, size = None):
+    def __fetchmany(self, size):
         if size is None:
             size = self.arraysize
-        if not self.active:
-            return Cursor.fetchmany(self, size)
-        self._source.execute("FETCH %d FROM %s" % (size, self.name))
-        return Cursor.fetchall(self)
+        if self.active:
+            self._source.execute("FETCH %d FROM %s" % (size, self.name))
+    def fetchmany(self, size = None):
+        self.__fetchmany(size)
+        # if we're a server side cursor, retrieve all the results from the fetch()
+        if self.active:
+            return Cursor.fetchall(self)
+        return Cursor.fetchmany(self, size)
+    def fetchmany_dict(self, size = None):
+        self.__fetchmany(size)
+        # if we're a server side cursor, retrieve all the results from the fetch()
+        if self.active:
+            return Cursor.fetchall_dict(self)
+        return Cursor.fetchmany_dict(self, size)
 
     def close(self):
         if self.active and self._source.valid:
@@ -228,36 +290,43 @@ class IterCursor(Cursor):
 class Database(object):
     def __init__(self, cnx):
         self.__cnx = cnx
-        cnx.execute("START TRANSACTION")
         # for prepared statement cache
         self.__cache = {}
 
+    def __del__(self):
+        if self.__cnx is not None:
+            self.close()
+        del self.__cache
+        del self.__cnx
+
     def close(self):
         # deallocate statements
-        self.rollback()
+        if self.__cnx.transaction in [TRANS_INERROR, TRANS_INTRANS]:
+            self.__cnx.execute("ROLLBACK")
         for src, name in self.__cache.itervalues():
             if not src.valid:
                 continue
             try:
                 self.__cnx.execute("DEALLOCATE %s" % (name,))
             except pgsql.ProgrammingError:
-                self.rollback()
+                self.__cnx.execute("ROLLBACK")
             src.close()
         self.__cache = {}
         self.__cnx.close()
+        self.__cnx = None
 
     # Postgresql autocommits everything not inside a transaction
     # block, so to simulate autocommit=off, we have to start new
     # transactions as soon as we flush them
     def commit(self):
         self.__cnx.execute("COMMIT")
-        self.__cnx.execute("START TRANSACTION")
 
     def rollback(self):
         self.__cnx.execute("ROLLBACK")
-        self.__cnx.execute("START TRANSACTION")
 
     def execute(self, query, *params):
+        # the database's .execute() method does not start transactions
+        # automatically
         if not len(params):
             ret = self.__cnx.execute(query)
         elif isinstance(params[0], (tuple, list)):
@@ -300,9 +369,20 @@ class Database(object):
 # connects to a database
 def connect(database = def_base, user = def_user, password = def_passwd,
             host = def_host, port = def_port, opt = def_opt, tty = def_tty):
+    global __CONNLIST
     # open the connection
     cnx = _pgsql.connect(database, user, password, host, port, opt, tty)
-    return Database(cnx)
+    db = Database(cnx)
+    __CONNLIST.append(weakref.ref(db))
+    return db
+
+def closeall():
+    global __CONNLIST
+    for ref in __CONNLIST:
+        db = ref()
+        if db is not None:
+            db.close()
+    __CONNLIST = []
 
 ### types handling
 
